/*
  Copyright (c) 2012, Matthias Schiffer <mschiffer@universe-factory.net>
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright notice,
       this list of conditions and the following disclaimer.
    2. Redistributions in binary form must reproduce the above copyright notice,
       this list of conditions and the following disclaimer in the documentation
       and/or other materials provided with the distribution.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/


#include "neigh.h"

#include <stdlib.h>


gp_babel_neigh_t* gp_babel_neigh_get(gmrf_iface_state_t *iface, const gmrf_addr_t *addr) {
	gp_babel_neigh_t *neigh = gp_babel_neigh_find(iface, addr);
	if (!neigh) {
		neigh = calloc(1, sizeof(gp_babel_neigh_t));
		neigh->next = iface->neighbours;
		neigh->iface = iface;
		iface->neighbours = neigh;
		neigh->last_packet = gmrf_time_unspec;
		neigh->last_hello = gmrf_time_unspec;
		neigh->last_ihu = gmrf_time_unspec;
		neigh->addr = *addr;

		neigh->txcost = GP_BABEL_INFINITY;
	}

	return neigh;
}

void gp_babel_neigh_unref_list(gp_babel_neigh_t *neigh) {
	gp_babel_neigh_t *next;
	for (; neigh; neigh = next) {
		next = neigh->next;

		neigh->iface = NULL;
		neigh->next = NULL;

		if (!neigh->ref)
			free(neigh);
	}
}

void gp_babel_neigh_unref(gp_babel_neigh_t *neigh) {
	if (!(--neigh->ref) && !neigh->iface)
		free(neigh);
}

uint16_t gp_babel_neigh_get_rxcost(gmrf_context_t *ctx, const gp_babel_neigh_t *neigh) {
	if (!neigh->hello_log || !neigh->hello_interval || !neigh->iface)
		return GP_BABEL_INFINITY;

	int timediff = gp_babel_since(ctx, neigh->last_hello);
	int shift = (timediff - neigh->hello_interval/2)/neigh->hello_interval;

	if (shift >= 16)
		return GP_BABEL_INFINITY;

	int received = __builtin_popcount((neigh->hello_log << shift) & 0xffff);

	if (received == 0)
		return GP_BABEL_INFINITY;
	else
		return (0x1000/received);
}

uint16_t gp_babel_neigh_get_txcost(gmrf_context_t *ctx, const gp_babel_neigh_t *neigh) {
	if (gp_babel_since(ctx, neigh->last_ihu) > GP_BABEL_IHU_TIMEOUT(neigh->ihu_interval) || !neigh->iface)
		return GP_BABEL_INFINITY;
	else
		return neigh->txcost;
}

uint16_t gp_babel_neigh_get_cost(gmrf_context_t *ctx, const gp_babel_neigh_t *neigh) {
	if (!neigh) /* self */
		return 0;

	uint16_t txcost = gp_babel_neigh_get_txcost(ctx, neigh);
	if (txcost < 256)
		txcost = 256;

	uint32_t cost = (txcost * gp_babel_neigh_get_rxcost(ctx, neigh)) >> 8;

	if (cost > GP_BABEL_INFINITY)
		return GP_BABEL_INFINITY;
	else
		return cost;
}

void gp_babel_neigh_reset(gmrf_context_t *ctx, gmrf_iface_state_t *iface, gp_babel_neigh_t *neigh) {
	gp_babel_send_route_request(ctx, iface, neigh, NULL);
}
